#------------------------------------------------------------------
# $Copyright: Copyright (c) 2010 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/AgentBuilder/AgentBuilder.pm,v 1.20 2010/11/29 05:39:46 lpatil Exp $
#
# $Author: lpatil $
# $Date: 2010/11/29 05:39:46 $
# $Revision: 1.20 $
#
# Provides:                            AgentBuilder
# Author:                              lspatil
# Creation Date:                       05 Sep, 2006
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package AgentBuilder;

my ( $CLUS_HOME ) = ();

#---------------------
# Setup CLUS_HOME..
#---------------------
if ( $ENV { CLUSTER_HOME } ) {
  $CLUS_HOME ||= $ENV{CLUSTER_HOME};
}
else { # VCS
  $CLUS_HOME ||= $ENV{VCS_HOME} ||= "/opt/VRTSvcs";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = $ENV{"HADECRYPT"} || "/opt/VRTSvcs/sbin/vcsdecrypt";

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: AgentBuilder.pm,v 1.20 2010/11/29 05:39:46 lpatil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

use constant EPOP                             => 10;
use constant SecondLevelMonitorTime           => 60;
use constant CHECK_INTERVAL                   => 10;



#--------------------------------
# Define AgentBuilder subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String StartProgram,
                              IN String StopProgram,
                              IN String CleanProgram,
                              IN String MonitorProgramFrequency,
                              IN String MonitorProgram,
                              IN String User,
                              IN String EnvFile,
                              IN String MonitorSequence,
                              IN String ListenAddressPort,
                              IN String Association PidFilesPatterns 
                              IN String Vector MonitorProcessPatterns   );


 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName                  Entry point name
 $localStartProgram                Start program for the AgentBuilder application 
 $localStopProgram                 Stop program for the AgentBuilder application 
 $localCleanProgram                Clean program for the AgentBuilder application 
 $localMonitorProgramFrequency     Cycle count to run monitor program
 $localMonitorProgram              External Monitor program for the AgentBuilder application 
 $localUser                        AgentBuilder application user
 $localEnvFile                     Environment File
 $localMonitorSequence             Decides order of monitoring methods   
 $localListenAddressPort           Listen Address and Port of AgentBuilder application Server
 $localPidFilesPatterns            Hash of PidFiles and Process pattern
 $localMonitorProcessPatterns      Perl Regex filter for processes
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localStartProgram,
       $localStopProgram,
       $localCleanProgram,
       $localMonitorProgramFrequency,
       $localMonitorProgram,
       $localUser,
       $localEnvFile,
       $localMonitorSequence,
       $localListenAddressPort,
       $localhrPidFilesPatterns,
       $localarMonitorProcessPatterns ) = @_;

  my $localPidFile;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName              [%s]
  StartProgram                [%s]
  StopProgram                 [%s]
  CleanProgram                [%s]
  MonitorProgramFrequency     [%s]
  MonitorProgram              [%s]
  User                        [%s]
  EnvFile                     [%s]
  MonitorSequence             [%s]
  ListenAddressPort           [%s] ",   
  $sMyName,   $sEntryPointName,   $localStartProgram,
  $localStopProgram,    $localCleanProgram,
  $localMonitorProgramFrequency, $localMonitorProgram,   $localUser,   
  $localEnvFile, $localMonitorSequence, $localListenAddressPort);

  while ( my ($key, $value) = each(%$localhrPidFilesPatterns) ) {

  &logTrace ( 20002, " <%s>
  PidFile           [%s]
  Pattern           [%s]",
  $sMyName, $key, $value );
  }

  foreach  my $key ( @$localarMonitorProcessPatterns)
  {
    &logTrace ( 20003, " <%s> MonitorProcessPatterns           [%s]", $sMyName, $key );
  }

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  my @temparr = @$localarMonitorProcessPatterns;
  my $inumProcesses =  $#temparr + 1   ;

  my $inumPidFilesPatterns =  scalar keys %$localhrPidFilesPatterns;

  if( ($inumPidFilesPatterns > 0) || ($inumProcesses > 0 ) || ( "" ne $localListenAddressPort) || ( "" ne $localMonitorProgram)  )
  {
    #------------------------------------------
    # At least one monitor attribute is defined.
    #------------------------------------------
  }
  else
  {
    #-----------------------------------------------------------
    # If none of the parameters are specified to monitor
    # the processes, return state will be set to STATE_UNKNOWN, 
    # Here we should return ArgsValid failed.
    #-----------------------------------------------------------
   
    &logError ( 20004, " <%s> None of the parameters (PidFilesPatterns, MonitorProcessPatterns, ListenAddressPort, MonitorProgram) 
                       are specified to monitor the Application. inumProcesses = [%s] inumPidFilesPatterns = [%s] ", 
                       $sMyName, $inumProcesses, $inumPidFilesPatterns);
   
    $iReturnCode = 0 ; # Failure..

  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  if(  "" ne $localMonitorSequence)
  {
    my @MonitorMethods = split(' ',$localMonitorSequence);
    my $MonitorMethod_defined_Count=0;
    my $bOnlyMonitorProgram = 0;

    foreach my $CurrentMonitorMethod (@MonitorMethods)
    {
       if( ( "ListenAddressPort" eq $CurrentMonitorMethod) || ( "MonitorProgram" ne $CurrentMonitorMethod) || ( "PidFilesPatterns" ne $CurrentMonitorMethod) || ( "MonitorProcessPatterns" ne $CurrentMonitorMethod)  )
       {
         if ( ( "ListenAddressPort" eq $CurrentMonitorMethod ) && ( "" ne $localListenAddressPort) )
         {
           $MonitorMethod_defined_Count++;
         } 
         if ( ( "MonitorProgram" eq $CurrentMonitorMethod ) && ( "" ne $localMonitorProgram)  && ( $localMonitorProgramFrequency > 0 ) )
         {
           # Do not allow "only" MonitorProgram to run with MonitorProgramFrequency > 1
           $bOnlyMonitorProgram = 1;
           $MonitorMethod_defined_Count++;
         } 
         if ( ( "PidFilesPatterns" eq $CurrentMonitorMethod ) && ($inumPidFilesPatterns > 0) )
         {
           $MonitorMethod_defined_Count++;
         } 
         if ( ( "MonitorProcessPatterns" eq $CurrentMonitorMethod ) && ($inumProcesses > 0 ) )
         {
           $MonitorMethod_defined_Count++;
         } 
       } 
       else
       {
          &logError ( 20005, " <%s>: Invalid <%s> monitoring method specified. Allowed set of monitoring methods MonitorProcessPatterns PidFilesPatterns ListenAddressPort MonitorProgram ", $sMyName, $CurrentMonitorMethod);
          $iReturnCode = 0 ; # Failure..
       } 
    } 
    if ( ( 1 == $MonitorMethod_defined_Count ) && $bOnlyMonitorProgram && ( $localMonitorProgramFrequency > 1 ) )
    {
          &logError ( 20006, " <%s>: MonitorProgramFrequency must be set to 1 if only monitoring method is MonitorProgram.", $sMyName);
          $iReturnCode = 0 ; # Failure..
    }
    if ( $MonitorMethod_defined_Count > 0 ) 
    {
          &logTrace ( 20007, " <%s>: At least one of the monitoring method specified as part of MonitorSequence has non-null value.", $sMyName);
    }
    else
    {
          &logError ( 20008, " <%s>: At least one of the monitoring method specified as part of MonitorSequence should have non-null value.", $sMyName);
          $iReturnCode = 0 ; # Failure..
    }
  }
  else
  {
      #--------------------------------
      # MonitorSequence can not be NULL
      #--------------------------------

      &logError ( 20009, " <%s>: MonitorSequence can not be NULL, Default MonitorSequence is given, user can alter MonitorSequence as per available monitoring methods.", $sMyName);
      $iReturnCode = 0 ; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------
  # Validate user..
  #-----------------------
  unless ( defined getpwnam $localUser ) {
    &logError ( 20010, "<%s> User name [%s] not valid.",
    $sMyName, $localUser );
    $iReturnCode = 0; # Failure..
  }

  #----------------------------
  # Validate environment file..
  #----------------------------
  if ( ( $sEntryPointName eq 'online' ) || ( $sEntryPointName eq 'offline' ) ) {
    if ( "" ne $localEnvFile ) { 
      if ( ! -e $localEnvFile ) {
        &logError ( 20011,
        "<%s> EnvFile [%s] does not exist", $sMyName, $localEnvFile );
        $iReturnCode = 0; # Failure..
      }
    }
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName    StartProgram     StopProgram     
       CleanProgram MonitorProgramFrequency MonitorProgram User EnvFile  ListenAddressPort PidFilesPatterns MonitorProcessPatterns);

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( StartProgram User  );
  my @lsOfflineArgs = qw ( StopProgram User  );
  my @lsMonitorArgs = qw ( User);
  my @lsCleanArgs   = qw ( User );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20012,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  if(  "" ne $localListenAddressPort)
  {
    my $sHost;
    my $sPort;
    ($sHost,$sPort) = split(/:/,$localListenAddressPort);
    if( (  "" ne $sHost) && (  "" ne $sPort ) )
    {
      #------------------------------------
      # Host and Port number are specified.
      #------------------------------------
    }
    else
    {
      &logError ( 20013, " <%s> Host and Port part must be specified. ", $sMyName); 
      $iReturnCode = 0 ; # Failure..
    }  
  }
  
  if ( $sEntryPointName eq 'online' ) {
    #--------------------------------------
    # MonitorProgram has to be executable..
    #--------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {                        # And not executable..
      &logError ( 20014,
      "<%s> MonitorProgram [%s] not executable", $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # CleanProgram has to be executable..
    #------------------------------------
    if ( $localCleanProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {                      # And not executable..
      &logError ( 20015,
      "<%s> CleanProgram [%s] not executable", $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20016,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String StartProgram,
                         IN String User,
                         IN String EnvFile );

 Description : Start logic needed for online entry point

 Parameters :

 $sStartProgram             Start program for the AgentBuilder application
 $sUser                     AgentBuilder application user
 $sEnvFile                  Environment File

 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sStartProgram, 
       $sUser, 
       $sEnvFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20017, "Subroutine <%s> called with:
  StartProgram  [%s]
  User          [%s]
  EnvFile       [%s]",
  $sMyName, $sStartProgram, $sUser, $sEnvFile );
  

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  my $iRunInBG    = 0;

  if ( $sStartProgram =~ /&\s*$/)
  {
    $iRunInBG      = 1;
    $sStartProgram =~ s/&//g ;
  }

  my ( $sCommand, $sCommandArgs) = $sStartProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..
  #------------------------------------------
  # Determine the expiration time of the EP.
  #------------------------------------------
  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };

  
  my $iRelevantTimeout = 0;
  $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
  if ( $iRelevantTimeout <= 0 )
  {
    &logError( 20018, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
    $iRelevantTimeout = 60;
  }
  my $iTillWhen = ( $^T + $iRelevantTimeout ) - AgentBuilder::EPOP;
  my $iTimeOut  = ($iTillWhen - time);

  my $iExitCode;
  my $bTimedOut;

  my @lsCommandOut;
  my $bSourceEnv   = 0;
  my $sOutFileName = "/tmp/.VRTS${sResourceName}_${sEntryPointName}_$$.out";
  my $sFromDir     = undef;

  if ( $iRunInBG )
  {
    $iReturnCode =Sys::RunWithEnvCmdInBG ( $sEnvFile,
                             $bSourceEnv,
                             $sCommand,
                             $sCommandArgs,
                             $iExitCode,
                             $sUser,
                             $sOutFileName,
                             \$bTimedOut,
                             $iTimeOut,
                             $sFromDir      );
  }
  else
  {
    $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iExitCode,
                                                 $sUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      );
  }


  if ( $iReturnCode )
  {
    #----------------------
    # Start program success
    #----------------------
    &logTrace ( 20019, "<%s>: Start command [%s] with args [%s] was successful",
                $sMyName, $sCommand, $sCommandArgs);
  }
  else
  {
    if ( $bTimedOut )
    {
      &logError ( 20020, "<%s>: Start command [%s] with args [%s] timed out",
                  $sMyName, $sCommand, $sCommandArgs);
    }
    else
    {
      &logError ( 20021, "<%s>: Start command [%s] with args [%s] failed to execute",
                  $sMyName, $sCommand, $sCommandArgs);
    }
    $iReturnCode =0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String StopProgram,
                        IN String User,
                        IN String EnvFile,
                        IN String Association PidFilesPatterns,
                        IN String Vector MonitorProcessPatterns);

 Description : Stop logic needed for offline entry point

 Parameters :

 $sStopProgram              Start program for the AgentBuilder application
 $sUser                     AgentBuilder application user
 $sEnvFile                  Environment File
 $hrPidFilesPatterns        Association of Absolute path to PID file of service and corresponding pattern
 $arMonitorProcessPatterns  Perl Regex filter for process
 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sStopProgram, 
       $sUser, 
       $sEnvFile,
       $hrPidFilesPatterns,
       $arMonitorProcessPatterns ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20022, "Subroutine <%s> called with:
  StopProgram   [%s]
  User          [%s]
  EnvFile       [%s]",
  $sMyName, $sStopProgram, $sUser, $sEnvFile  );

  while ( my ($key, $value) = each(%$hrPidFilesPatterns) ) {
          &logTrace ( 20028, " <%s>
                      PidFile [%s] => Pattern [%s]",
                      $sMyName, $key, $value );
  }

 foreach  my $key ( @$arMonitorProcessPatterns)
 {
         &logTrace ( 20029, " <%s>
                     MonitorProcessPatterns [%s]",
                     $sMyName, $key );
 }

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  my ( $sCommand, $sCommandArgs) = $sStopProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..
  #------------------------------------------
  # Determine the expiration time of the EP.
  #------------------------------------------
  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my $iRelevantTimeout = 0;
  $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
  if ( $iRelevantTimeout <= 0 )
  {
    &logError( 20023, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
    $iRelevantTimeout = 60;
  }
  my $iTillWhen = ( $^T + $iRelevantTimeout ) - AgentBuilder::EPOP;
  my $iTimeOut  = ($iTillWhen - time);

  my $iExitCode;
  my $bTimedOut;

  my @lsCommandOut;
  my $bSourceEnv   = 0;
  my $sOutFileName = undef;
  my $sFromDir     = undef;

  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 $iExitCode,
                                                 $sUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
  {
    #----------------------
    # Stop program success
    #----------------------
    &logTrace ( 20024, "<%s>: Stop command [%s] with args [%s] was successful",
                $sMyName, $sCommand, $sCommandArgs);
  }
  else
  {
    if ( $bTimedOut )
    {
      &logError ( 20025, "<%s>: Stop command [%s] with args [%s] timed out",
                  $sMyName, $sCommand, $sCommandArgs);
    }
    else
    {
      &logError ( 20026, "<%s>: Stop command [%s] with args [%s] failed to run",
                  $sMyName, $sCommand, $sCommandArgs);
    }
 
    $iReturnCode = 0;
  }

  $iTimeOut  = ($iTillWhen - time); # Time still remaining in offline EP

  while ( $iTimeOut > CHECK_INTERVAL )
  {
    my @liPids = ();
    my @liPids_PidFilesPatterns = ();

    my @liPidPatterns = ();
    my @liOnlinePidPatterns = ();
  
    my @liPids_MonitorProcessPatterns = ();

    my @temparr = @$arMonitorProcessPatterns;
    my $inumProcesses =  $#temparr + 1   ;
    my $inumPidFilesPatterns =  scalar keys %$hrPidFilesPatterns;

    if ( $inumPidFilesPatterns > 0) 
    {
       AgentBuilder::GetPidFileState(\@liPids_PidFilesPatterns, $sUser,  $sEnvFile, $hrPidFilesPatterns, );
   
       while ( my ($key, $value) = each(%$hrPidFilesPatterns) )
       {
         if( $value ne "" )
         {
           push ( @liPidPatterns, $value ); # Push..
         }
       }
       AgentBuilder::GetProcessState(\@liOnlinePidPatterns, $sUser,  $sEnvFile, \@liPidPatterns, 0 );
  
       @liPids_PidFilesPatterns = (@liPids_PidFilesPatterns, @liOnlinePidPatterns);
   
    }
    if( $inumProcesses > 0 )  
    {
       AgentBuilder::GetProcessState(\@liPids_MonitorProcessPatterns, $sUser,  $sEnvFile, $arMonitorProcessPatterns, 0 );
    }
    @liPids = (@liPids_PidFilesPatterns, @liPids_MonitorProcessPatterns);

    if ( @liPids < 1 )
    {
      &logTrace ( 20150, "<%s>: Application processes not running", $sMyName);
      return $iReturnCode;
    }
    else
    {
      &logTrace ( 20151, "<%s>: Application processes still running. Sleeping for 10 seconds", $sMyName);
      sleep CHECK_INTERVAL;
      $iTimeOut  = ($iTillWhen - time); 
    }
  }

  if ( $iTimeOut <= CHECK_INTERVAL )
  {
    &logInfo ( 20152, "<%s>: Not enough time left to do wait for application process checks", $sMyName);
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String CleanProgram,
                        IN String User,
                        IN String EnvFile,
                        IN String Association PidFilesPatterns,
                        IN String Vector MonitorProcessPatterns);

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sCleanProgram             Clean program for the AgentBuilder application
 $sUser                     AgentBuilder application user
 $sEnvFile                  Environment File
 $hrPidFilesPatterns        Association of Absolute path to PID file of service and corresponding pattern
 $arMonitorProcessPatterns  Perl Regex filter for process
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCleanProgram,
       $sUser, 
       $sEnvFile,
       $hrPidFilesPatterns,
       $arMonitorProcessPatterns ) = @_;

  my $sPidFile;
 
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20027, "Subroutine <%s> called with:
  CleanProgram             [%s]
  User                     [%s]
  EnvFile                  [%s]",
  $sMyName, $sCleanProgram, $sUser, $sEnvFile );

  while ( my ($key, $value) = each(%$hrPidFilesPatterns) ) {
          &logTrace ( 20028, " <%s>
                          PidFile [%s] => Pattern [%s]",
                          $sMyName, $key, $value );
  }

 foreach  my $key ( @$arMonitorProcessPatterns)
 {
         &logTrace ( 20029, " <%s>
                         MonitorProcessPatterns           [%s]",
                         $sMyName, $key );
 }

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;
  my @liPids = ();
  my @liPids_PidFilesPatterns = ();

  my @liPidPatterns = ();
  my @liOnlinePidPatterns = ();

  my @liPids_MonitorProcessPatterns = ();

  my @temparr = @$arMonitorProcessPatterns;
  my $inumProcesses =  $#temparr + 1   ;
  my $inumPidFilesPatterns =  scalar keys %$hrPidFilesPatterns;

  if ( $inumPidFilesPatterns > 0) 
  {
     AgentBuilder::GetPidFileState(\@liPids_PidFilesPatterns, $sUser,  $sEnvFile, $hrPidFilesPatterns, );
 
     while ( my ($key, $value) = each(%$hrPidFilesPatterns) )
     {
       if( $value ne "" )
       {
         push ( @liPidPatterns, $value ); # Push..
       }
     }
     AgentBuilder::GetProcessState(\@liOnlinePidPatterns, $sUser,  $sEnvFile, \@liPidPatterns, 0 );

     @liPids_PidFilesPatterns = (@liPids_PidFilesPatterns, @liOnlinePidPatterns);
 
  }
  if( $inumProcesses > 0 )  
  {
     AgentBuilder::GetProcessState(\@liPids_MonitorProcessPatterns, $sUser,  $sEnvFile, $arMonitorProcessPatterns, 0 );
  }
  @liPids = (@liPids_PidFilesPatterns, @liPids_MonitorProcessPatterns);

  if( $sCleanProgram =~ /(\S+)/ ) {
    my ( $sCommand, $sCommandArgs) = $sCleanProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

    #------------------------------------------
    # Determine the expiration time of the EP.
    #------------------------------------------
    my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
    my $iRelevantTimeout = 0;
    $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
    if ( $iRelevantTimeout <= 0 )
    {
      &logError( 20030, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
      $iRelevantTimeout = 60;
    }
    my $iTillWhen = ( $^T + $iRelevantTimeout ) - AgentBuilder::EPOP;
    my $iTimeOut  = $iTillWhen - time;


    my $iExitCode;
    my $bTimedOut;

    my @lsCommandOut;
    my $bSourceEnv   = 0;
    my $sOutFileName = undef;
    my $sFromDir     = undef;


    if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                   $bSourceEnv,
                                                   $sCommand,
                                                   $sCommandArgs,
                                                   \$iExitCode,
                                                   $sUser,
                                                   \@lsCommandOut,
                                                   \$bTimedOut,
                                                   $iTimeOut,
                                                   $sOutFileName,
                                                   $sFromDir      ) )

    {
      #----------------------
      # Clean program success
      #----------------------
      &logTrace ( 20031, "<%s>: Clean command [%s] with args [%s] was successful",
                $sMyName, $sCommand, $sCommandArgs);
    }
    else
    {
      $iReturnCode = 0;
      if ( $bTimedOut )
      {
        #-------------------------
        # Clean program timed out.
        #-------------------------
        &logError ( 20032, "<%s>: Clean command [%s] with args [%s] timed out ",
                  $sMyName , $sCommand , $sCommandArgs);
      }
      else
      {
        #---------------------------------------------
        # Clean program did not execute successfully.
        #---------------------------------------------
        &logError ( 20033, "<%s>: Clean command [%s] with args [%s] failed to run",
                  $sMyName , $sCommand , $sCommandArgs);
      }
    }
  }
  #---------------------------
  # Kill AgentBuilder Processes..
  #---------------------------

  foreach my $iTmpPid (@liPids)
  {
    #-------------------------------
    # Send TERM in the first round.
    #-------------------------------
    if (kill 0, $iTmpPid)
    {
      &logInfo ( 20034 , "<%s>: Sending SIGTERM to pid [%s]",
      $sMyName , $iTmpPid); 
      kill 15, $iTmpPid;
    }
  }
  foreach my $iTmpPid (@liPids)
  {
    #----------------------------------------------------------
    # Send KILL if the processes are still found to be running
    #----------------------------------------------------------
    if (kill 0, $iTmpPid)
    {
      &logInfo ( 20035 ,"<%s>: Sending SIGKILL to pid [%s]",
                  $sMyName , $iTmpPid );
      kill 9, $iTmpPid;
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int GetPidFileState (  OUT List                liOnlinePids ,
                                   IN String               User,
                                   IN String               EnvFile
                                   IN String Hash          PidFilesPatterns );

 Description: Performs the first level monitor checks

 Parameters :

 $rliOnlinePids             Reference to list of Online PID hash
 $sUser                     AgentBuilder application user
 $sEnvFile                  Environment File
 $arPidFilesPatterns        Hash of Application Pid files and corresponding pattern for process in the Pid file 


 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetPidFileState {

  #----------------
  # Get Arguments..
  #----------------
  my (  $rliOnlinePids,
        $sUser,
        $sEnvFile,
        $hrPidFilesPatterns  ) = @_;

  my $sPidFile;
  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20036, "Subroutine <%s> called with:
  User                     [%s]
  EnvFile                  [%s] ",
  $sMyName, $sUser, $sEnvFile );

  while ( my ($key, $value) = each(%$hrPidFilesPatterns) ) {
          &logTrace ( 20037, " <%s>
                          PidFile [%s] => Pattern [%s]",
                          $sMyName, $key, $value );
  }

  #--------------------------------
  # Setup Pidfile processes pattern list
  #--------------------------------

  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my @sMonitorList;

  my $iOnlineCount  = 0;
  my $iOfflineCount = 0;
  my ( $rhlProc, @lProcessHash ) = ();

  my ( $iProcessCount, $iFilteredProcCount, $iFound, 
       $iPid, $sLang, $sKey, $sProc, $sTemp ) = ();

  @$rliOnlinePids = ();          # Reset..


  #------------------------------------------------------------
  # Open the PidFile and check whether the process is running.
  #------------------------------------------------------------



   while ( my ($key, $value) = each(%$hrPidFilesPatterns) ) 
        {

     $sPidFile = $key;
     if ( -f $sPidFile)
     {
        if( open ( PIDFILE,"$sPidFile") )
        {
           ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
            s/.*?(\d+)$/$1/  # Get PID..
         } <PIDFILE>;

         $iPid = '' unless defined $iPid;      # Avoid undefs..

         close ( PIDFILE ) ||                  # Close file..
         &logWarn ( 20038,
            "<%s> Could not close PID file [%s] after read [%s]",
            $sMyName, $sPidFile, $! );


         #------------------------
         # Check if PID is valid..
         #------------------------
         if ( $iPid =~ /\d+/ &&     # If the valid number..
               kill ( 0, $iPid ) ) 
         {
             #------------------------
             # Is a PID that's alive..
             #------------------------

             &logTrace ( 20039,
                   "<%s> PID [%s] from PID file [%s] exists",
                   $sMyName, $iPid, $sPidFile );

             #----------------------------------------------------------
             # Now Pid is there, check for corresponding process pattern
             #----------------------------------------------------------
             if( $value ne "" )
             {
              my $sFilter = $value;
              my $iNumberOfMatches = 0;

                  #-----------------------
                  # Search in proc table..
                  #-----------------------
              if ( AgentBuilder::GetPidDetails ( $iPid,
                       \@lProcessHash ) ) 
              {
                       #--------------
                       # Got details..
                       #--------------
                       $iNumberOfMatches = 0;

                #------------------------------------------------
                # Spin the list of processes we need to monitor
                #------------------------------------------------

                foreach $rhlProc ( @lProcessHash ) 
                {
                  &logTrace ( 20040,
                        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
                        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
                        $$rhlProc { COMMAND } );

                  #--------------
                  # Now compare..
                  #--------------

                  if ( $$rhlProc { COMMAND } =~ /$sFilter/ ) 
                  {
                     #----------------
                     # Filter match..
                     #----------------
                     push ( @$rliOnlinePids, $iPid ); # Push..
                     &logTrace ( 20041,
                           "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
                           $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sFilter);
                     $iNumberOfMatches = $iNumberOfMatches + 1;
                  }
                  else 
                  {
                     &logTrace ( 20042,
                           "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
                           $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sFilter);
                  }
                }
              }
              else 
              {
                 &logTrace( 20043, "<%s>: Could not get details for PID [%s]",$sMyName, $iPid);
                 $iOfflineCount++;
              }

              &logTrace( 20044, "<%s>: Number of PIDs Matched [%s] ",
                    $sMyName, $iNumberOfMatches);

              if ($iNumberOfMatches < 1  )
              {
                 &logError( 20045, "<%s>: Pid [%s] is present but process with corresponding pattern [%s] is not running",
                       $sMyName, $iPid, $sFilter);
                 $iOfflineCount++;
              }
              else
              {
                 &logTrace( 20046, "<%s>: Pid [%s] is present and process with corresponding pattern [%s] is  running",
                       $sMyName, $iPid, $sFilter);
                 $iOnlineCount++;
                
              }
             }
             else
             {
                &logTrace( 20047, "<%s>: Pid [%s] is present ",
                                  $sMyName, $iPid);
                $iOnlineCount++;
                push ( @$rliOnlinePids, $iPid ); # Push..
             }
         }
         else
         {
            &logTrace( 20048, "<%s>: Pid [%s] from PidFile [%s] not running",$sMyName, $iPid, $sPidFile);
            $iOfflineCount++;
         }
         
        }
        else
        {
           &logError( 20049, "<%s>: Unable to open PidFile [%s]",$sMyName, $sPidFile);
           $iOfflineCount++;
        }
     }
     else
     {
        &logTrace( 20050, "<%s>: PidFile [%s] does not exist or is not a regular file",$sMyName, $sPidFile);
        $iOfflineCount++;
     }
   }
  
 
  my $iReturnState=VCSDef_MonitorExit::ONLINE;

  if (( scalar keys %$hrPidFilesPatterns ) > 0 )
  {

     if ( $iOfflineCount > 0 )
     {
        if (   $sEntryPointName ne 'monitor'  &&
               ( $iOnlineCount > 0 ) )
        {
           $iReturnState = VCSDef_MonitorExit::PARTIAL_ONLINE;
        }
        else
        {
           $iReturnState = VCSDef_MonitorExit::OFFLINE;
        }
     }
     else
     {
        if($iOnlineCount > 0)
        {
           $iReturnState = VCSDef_MonitorExit::ONLINE;
        }
        else
        {
           $iReturnState = VCSDef_MonitorExit::OFFLINE;
        }
     }
  }
  &logTrace(20051, "<%s>:Only PidFilesPatterns check done. OfflineCount [%s] OnlineCount [%s] ReturnCode [%s]",
        $sMyName, $iOfflineCount, $iOnlineCount, $iReturnState);
  return $iReturnState;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcessState (  OUT List                liOnlinePids ,
                                   IN String               User,
                                   IN String               EnvFile
                                   IN String Vector        MonitorProcessPatterns
                                   IN Boolean              UseFFLM  );

 Description: Performs the first level monitor checks

 Parameters :

 $rliOnlinePids             Reference to list of Online PID hash
 $sUser                     AgentBuilder application user
 $sEnvFile                  Environment File
 $arMonitorProcessPatterns  Perl Regex filter for processes
 $bUseFFLM                  Flag to indicate if fast FLM is to be attempted


 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut


sub GetProcessState {

  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sUser, 
       $sEnvFile,
       $arMonitorProcessPatterns,
       $bUseFFLM  ) = @_;

  my $sPidFile; 

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20052, "Subroutine <%s> called with:
  User                     [%s]
  EnvFile                  [%s]
  UseFFLM                  [%s]",
  $sMyName, $sUser, $sEnvFile, $bUseFFLM  );

  foreach  my $key ( @$arMonitorProcessPatterns)
  {
    &logTrace ( 20053, " <%s>
    MonitorProcessPatterns           [%s]",
    $sMyName, $key );
  }

  #--------------------------------
  # Setup monitor processes list
  #--------------------------------

  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my @sMonitorList;

  my $iOnlineCount  = 0;
  my $iOfflineCount = 0;
  my $iPerlPid = $$; # Get current PID..
  my ( $rhlProc ) = ();

  my ( $iProcessCount, $iFilteredProcCount, $iUseRFLM, $iFound, 
       $iPid, $iUid, $sLang, $sKey, $sProc, $sTemp, 
       @lsUpdates, @lsValues,
       %hiProcPids, %hsApplicationDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();          # Reset..

  my $iReturnState = VCSDef_MonitorExit::ONLINE;

  #----------
  # Get UID..
  #----------
  $iUid = getpwnam $sUser; # Should not fail..


  my @hProcListHash = ();


  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( AgentBuilder::GetApplicationDetails ( $arMonitorProcessPatterns,
                                             \%hsApplicationDetails ) ) 
  {

    &logTrace ( 20054,
    "<%s> Got details for monitoring.", $sMyName );

    %hsCopy = %hsApplicationDetails;          # Backup..
    $iUseRFLM = scalar keys %hsCopy;       # Set to entire count..
  }
  else 
  {
    &logError ( 20055,
    "<%s> Failed to get details for monitoring", $sMyName );
    return VCSDef_MonitorExit::UNKNOWN;
  }

  #---------------------------------
  # Check if we attempt to be fast..
  #---------------------------------
  if ( $bUseFFLM ) { # FFLM..
    $iUseRFLM = 0;   # Hopefully skip..

    foreach $sKey ( keys %hsCopy ) {          # Use copy..
      @lsUpdates = ();                        # Reset..
      @lsValues = @{ $hsCopy { $sKey } };     # Get values..
      $sTemp .= "\n Application process [$sKey]\n"; # Display string..

      #----------------
      # Read in pairs..
      #----------------
      while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

        #-----------
        # Run FFLM..
        #-----------
        $iFound = AgentBuilder::FFLMStatus ( $sPidFile,
                                       $sProc,
                                       \$iPid );
        #----------------
        # Update status..
        #----------------
        push ( @lsUpdates, $sPidFile, $sProc, $iFound );

        $sTemp .= "PID File       [$sPidFile]\n";
        $sTemp .= "Search Process [$sProc]\n";
        $sTemp .= "FFLM Status    [$iFound]\n\n";
      }

      @{ $hsCopy { $sKey } } = @lsUpdates;

      #------------------------
      # Handle monitor status..
      #------------------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        push ( @$rliOnlinePids, $iPid ); # Push..
        $hiProcPids { $sKey } = $iPid;   # Store..
      }
      else {
        ++ $iUseRFLM; # Mark it for RFLM..
      }
    }

    #----------------
    # Print summary..
    #----------------
    &logTrace ( 20056,
    "<%s> Application  process details ---%s\n\n",
    $sMyName,  $sTemp );

    ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
  }
  else 
  { # Use RLFM..
    #-----------------------------------------
    # Mark status of each service as offline..
    #-----------------------------------------
    $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
    foreach $sKey ( keys %hsCopy ) 
    {
      @lsUpdates = (); # Reset..

      @lsValues = @{ $hsCopy { $sKey } }; # Get values..
      while ( scalar @lsValues ) 
      {   
        push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
      }

      @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
    }
  }


 #-----------------------
 # Regular slow :-( FLM..
 #-----------------------
 if ( $iUseRFLM ) 
 {
   &logTrace ( 20057,
   "<%s> Running regular first level monitor", $sMyName );

   #----------------------------------
   # Set LANG for better performance..
   #----------------------------------
   $sLang = $ENV { LANG }; # Backup..
   $ENV { LANG } = 'C';    # Now set..

   #--------------------------------
   # Get Entire process snap shot..
   #--------------------------------
   $iProcessCount = Proc::GetLongProcessListHash(\@hProcListHash);

   #----------
   # Restore..
   #----------
   defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

   #----------------------
   # Check if we got any..
   #----------------------
   if ( $iProcessCount ) {
     &logTrace ( 20058,
     "<%s> Got [%s] processes from proc table for filtering",
     $sMyName, $iProcessCount );
   }
   else {
     &logError ( 20059,
     "<%s> Call to get process list did not return any processes",
     $sMyName );
     return VCSDef_MonitorExit::OFFLINE; # Offline..
   }

   #-----------------------
   # Search for processes..
   #-----------------------
   foreach $rhlProc ( @hProcListHash ) {
     #--------------------------------------------------------------
     # Each $rhlProc is a ref. to a hash with the following keys..
     # "UID"     =>  <user id>
     # "PID"     =>  <process id>
     # "PPID"    =>  <parent process id>
     # "COMMAND" =>  <command string>
     #--------------------------------------------------------------

     &logTrace ( 20060, "<%s> Searching PID [%s] UID [%s] CMD [%s]",
     $sMyName, $$rhlProc { PID }, $$rhlProc { UID },
     $$rhlProc { COMMAND } );

     next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
     $iPid = $$rhlProc { PID };             # Assign..

     #-----------------------------------------------
     # Look for processes that were offline in FFLM..
     #-----------------------------------------------
     if ( $iUid == $$rhlProc { UID } ) {
       foreach $sKey ( keys %hsCopy ) {
         @lsUpdates = (); # Reset..
         @lsValues = @{ $hsCopy { $sKey } }; # Get values..

         #--------------------------------------------------
         # Should have details per three elements of array..
         # <PIDFILE> <PROCESS> <STATE>
         #--------------------------------------------------
         while ( ( $sPidFile, $sProc, $iFound ) =
                   splice ( @lsValues, 0, 3 ) ) {
           if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {          # Now found..

             if ( $iPid ne $iPerlPid ) {

                 unshift ( @$rliOnlinePids, $iPid );   # Store..

               if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                 &logTrace ( 20061,
                 "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                 $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                 $sProc );

                 $hiProcPids { $sKey } = $iPid;        # Store..
                 $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
               }
             }
             else {
               &logTrace ( 20062,
               "<%s> Skipping self. PID [%s] PPID [%s] CMD [%s] with filter [%s]",
               $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
               $sProc );
             }
           }

           #----------------
           # Update status..
           #----------------
           push ( @lsUpdates, $sPidFile, $sProc, $iFound );
         }

         @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
       }
     }
   }
 }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline) = ();

  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {

    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20063,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( AgentBuilder::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20064,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20065,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }
      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnState = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN; # Verify..

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if (   $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnState )  &&
         scalar @$rliOnlinePids ) 
  {
    $iReturnState = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20066,
    "<%s> Found [%s] PIDs [%s] 
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ), $iReturnState );
  }


  return $iReturnState;

}

=pod

 -------------------------------------------------------------------------
 Function : int SequencedMonitor ( IN String Vector MonitorMethods
                                   IN String        User,
                                   IN String        EnvFile,
                                   IN String Hash   PidFilePatterns,
                                   IN String Vector MonitorProcessPatterns,
                                   IN String        ListenAddressPort,
                                   IN String        MonitorProgram, 
                                   IN Integer       MonitorProgramFrequency,
                                   IN Integer       RelevantTimeout, 
                                   IN Integer       numProcesses,
                                   IN Integer       numPidFilesPatterns,
                                   IN String        ResourceName 
                                 );

 Description: Monitor the resource using specified monitoring methods in MonitorSequence attribute.

 Parameters :
 $arMonitorMethods         Vector of Monitor methods
 $sUser                    AgentBuilder application user
 $sEnvFile                 Environment File
 $hrPidFilesPatterns       Association of Absolute path to PID file of service and corresponding pattern
 $arMonitorProcessPatterns Vector of process strings to search for in proc table
 $sListenAddressPort       Host and Port address for socket test 
 $sMonitorProgram          Monitor program for the AgentBuilder application
 $iMonitorProgramFrequency Cycle count to run monitor program
 $iRelevantTimeout         Timeout value 
 $inumProcesses            Number of process patterns configured for the resource
 $inumPidFilesPatterns     Number of PidFile patterns configured for the resource
 $sResourceName            Name of the resource

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SequencedMonitor {

  #----------------
  # Get Arguments..
  #----------------

  my (  $arMonitorMethods,
        $sUser,
        $sEnvFile,
        $hrPidFilesPatterns,
        $arMonitorProcessPatterns,
        $sListenAddressPort,
        $sMonitorProgram,
        $iMonitorProgramFrequency,
        $iRelevantTimeout,
        $inumProcesses,
        $inumPidFilesPatterns, 
        $sResourceName ) = @_ ;

  my $AgentBuilder_Status;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20067, "Subroutine <%s> called", $sMyName );

  foreach my $CurrentMonitorMethod (@$arMonitorMethods)
  {
    
     #--------------------------------
     # Start monitoring the resource..
     #--------------------------------

     if ( $CurrentMonitorMethod eq "PidFilesPatterns" ) 
     {
         #---------------------------------------------------------
         # Start monitoring the resource using PidFilesPatterns...
         #---------------------------------------------------------
         if( $inumPidFilesPatterns > 0)
         {
            my @liDummyPids = ();
            $AgentBuilder_Status = AgentBuilder::GetPidFileState ( \@liDummyPids, $sUser, $sEnvFile, $hrPidFilesPatterns );
 
            #-----------------------------------------------------
            # Exit if, not online after PidFilesPatterns check..
            #-----------------------------------------------------
            if ( VCSDef_MonitorExit::ONLINE == $AgentBuilder_Status ) {
               &logTrace ( 20068,
                              "<%s> Resource [%s] online with state [%s] after PidFilesPatterns check.",
                              $sMyName, $sResourceName, $AgentBuilder_Status );
            }
            else {
               &logTrace ( 20069,
                              "<%s> Resource [%s] NOT online with state [%s] after PidFilesPatterns check.",
                              $sMyName, $sResourceName, $AgentBuilder_Status );
               return $AgentBuilder_Status;
            }
         }
     }
     elsif ( $CurrentMonitorMethod eq "MonitorProcessPatterns" ) 
     {
         #--------------------------------
         # Start monitoring the resource using MonitorProcessPatterns...
         #--------------------------------

         if( $inumProcesses > 0 )
         {
             my @liDummyPids = ();
             $AgentBuilder_Status = AgentBuilder::GetProcessState ( \@liDummyPids, $sUser, $sEnvFile, $arMonitorProcessPatterns, 1 );

             #------------------------------------------------
             # Exit if, not online after MonitorProcessPatterns check..
             #------------------------------------------------

             if ( VCSDef_MonitorExit::ONLINE == $AgentBuilder_Status ) {
                 &logTrace ( 20070,
                                   "<%s> Resource [%s] online with state [%s] after MonitorProcessPatterns check.",
                                   $sMyName, $sResourceName, $AgentBuilder_Status );
                 }
             else {
                 &logTrace ( 20071,
                                   "<%s> Resource [%s] NOT online with state [%s] after MonitorProcessPatterns check.",
                                   $sMyName, $sResourceName, $AgentBuilder_Status );
                 return $AgentBuilder_Status;
             }
         }
     }
     elsif ( $CurrentMonitorMethod eq "ListenAddressPort" ) 
     {
         #--------------------------------
         # Start monitoring the resource using ListenAddressPort...
         #--------------------------------

         if(  "" ne $sListenAddressPort   )
         {
             my $sHost;
             my $sPort;
             ($sHost,$sPort) = split(/:/,$sListenAddressPort);
             $AgentBuilder_Status = AgentBuilder::SocketTest(   $sUser, $sEnvFile, $sHost, $sPort );

             if ( VCSDef_MonitorExit::ONLINE ==  $AgentBuilder_Status )
             {
                 &logTrace ( 20072, "<%s>: Socket connection test -> Instance is online",$sMyName);
             }
             else
             {
                 &logTrace ( 20073, "<%s>: Socket connection test -> Instance is not online",$sMyName);
                 return $AgentBuilder_Status;
             }

         }

     }
     elsif ( $CurrentMonitorMethod eq "MonitorProgram" ) 
     {
         #--------------------------------
         # Start monitoring the resource using MonitorProgram ...
         #--------------------------------

         if(  "" ne $sMonitorProgram && ( $iMonitorProgramFrequency > 0 ) )
         {
             $AgentBuilder_Status = AgentBuilder::Monitor ( $sMonitorProgram,
                                                            $iMonitorProgramFrequency,
                                                            $sUser,
                                                            $sEnvFile,
                                                            $iRelevantTimeout);

             if ( VCSDef_MonitorExit::ONLINE ==  $AgentBuilder_Status )
             {
               #----------------------------------------------------------
               # message already logged as part of ExternalMonitorProgram.
               #----------------------------------------------------------
             }
             else
             {
               return $AgentBuilder_Status;
             }

         }

      }
      else
      {
        &logError( 20074, "<%s> Unknown Monitoring method <%s> specified in value of MonitorSequence attribute.", $sMyName, $CurrentMonitorMethod );
      }

} #for

return $AgentBuilder_Status;
}


=pod

-------------------------------------------------------------------------
Function : int HandlePartialOnline ( IN String        User,
                                     IN String        EnvFile,
                                     IN String Vector MonitorProcessPatterns,
                                     IN Integer       numProcesses,
                                     IN String Hash   PidFilePatterns,
                                     IN Integer       numPidFilesPatterns,
                                     IN String        ListenAddressPort, 
                                     IN String        CleanProgram
                                     );

Description: If the resource is in PARTIAL_ONLINE state and clean up the system before proper Online. 

Parameters :

$sUser                    AgentBuilder application user
$sEnvFile                 Environment File
$arMonitorProcessPatterns Vector of process strings to search for in proc table
$inumProcesses            Number of process patterns configured for the resource
$hrPidFilesPatterns       Association of Absolute path to PID file of service and corresponding pattern
$inumPidFilesPatterns     Number of PidFile patterns configured for the resource
$sCleanProgram            Clean program for the AgentBuilder application

Returns : Integer value as defined below:
100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
-------------------------------------------------------------------------

=cut


sub HandlePartialOnline {

  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $arMonitorProcessPatterns,
       $inumProcesses,
       $hrPidFilesPatterns,
       $inumPidFilesPatterns,
       $sListenAddressPort,
       $sCleanProgram )= @_ ;

  my $iOnlineCount  = 0;
  my $iOfflineCount = 0;
  my $iUnknownCount = 0;
  my $iPartialOnlineCount  = 0;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20075, "Subroutine <%s> called", $sMyName );

  my $iReturnCode = VCSDef_MonitorExit::OFFLINE;

  #------------------------------------------------------
  # Check for Process list for  PARTIAL_ONLINE state
  #------------------------------------------------------


  my ( $sCommand, $sCommandArgs) = $sCleanProgram # Separate
                                 =~ /(\S+)\s*(.*)/; # Cmd and args..

  #------------------------------------------
  # Determine the expiration time of the EP.
  #------------------------------------------
  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my $iRelevantTimeout = 0;
  $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
  if ( $iRelevantTimeout <= 0 )
  {
    &logError( 20076, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
    $iRelevantTimeout = 60;
  }
  my $iTillWhen = ( $^T + $iRelevantTimeout ) - AgentBuilder::EPOP;
  my $iTimeOut  = $iTillWhen - time;

  my $iExitCode;
  my $bTimedOut;

  my @lsCommandOut;
  my $bSourceEnv   = 0;
  my $sOutFileName = undef;
  my $sFromDir     = undef;

  my @liPids = ();
  my @liCopyProcessPids = ();

  if( $inumProcesses > 0 )
  {
    $iReturnCode = AgentBuilder::GetProcessState ( \@liPids, $sUser, $sEnvFile, $arMonitorProcessPatterns, 0 );
    @liCopyProcessPids = @liPids;
    if ( VCSDef_MonitorExit::PARTIAL_ONLINE == $iReturnCode )
    {
      $iPartialOnlineCount++;
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
      $iOfflineCount++;
    }
    elsif ( VCSDef_MonitorExit::ONLINE == $iReturnCode )
    {
      $iOnlineCount++;
    } 
    elsif ( VCSDef_MonitorExit::OFFLINE == $iReturnCode )
    {
      $iOfflineCount++;
    } 
    elsif ( VCSDef_MonitorExit::UNKNOWN == $iReturnCode )
    {
      $iUnknownCount++;
    } 
  }

  #------------------------------------------------------------
  # Check for PidFilePatterns list for  PARTIAL_ONLINE state
  #------------------------------------------------------------

  @liPids = ();
  my @liPidPatterns = ();
  my @liOnlinePidPatterns = ();
  my @liCopyPidFilePids = ();

  if( $inumPidFilesPatterns > 0)
  {
       $iReturnCode = AgentBuilder::GetPidFileState ( \@liPids, $sUser, $sEnvFile, $hrPidFilesPatterns );

       while ( my ($key, $value) = each(%$hrPidFilesPatterns) )
       {

          if( $value ne "" )
          {
            push ( @liPidPatterns, $value ); # Push..
          }
       }
       AgentBuilder::GetProcessState(\@liOnlinePidPatterns, $sUser,  $sEnvFile, \@liPidPatterns, 0 );

       @liPids = (@liPids, @liOnlinePidPatterns);
       @liCopyPidFilePids = @liPids;

       if ( VCSDef_MonitorExit::PARTIAL_ONLINE == $iReturnCode )
       {
          $iPartialOnlineCount++;
          $iReturnCode = VCSDef_MonitorExit::OFFLINE;
          $iOfflineCount++;
       }
       elsif ( VCSDef_MonitorExit::ONLINE == $iReturnCode )
       {
          $iOnlineCount++;
       }
       elsif ( VCSDef_MonitorExit::OFFLINE == $iReturnCode )
       {
          $iOfflineCount++;
       }
       elsif ( VCSDef_MonitorExit::UNKNOWN == $iReturnCode )
       {
          $iUnknownCount++;
       }

  }
 
  if(  "" ne $sListenAddressPort   )
  {
         my $sHost;
         my $sPort;
         ($sHost,$sPort) = split(/:/,$sListenAddressPort);
         $iReturnCode = AgentBuilder::SocketTest(   $sUser, $sEnvFile, $sHost, $sPort );

         if ( VCSDef_MonitorExit::ONLINE == $iReturnCode )
         {
           $iOnlineCount++;
         }
         elsif ( VCSDef_MonitorExit::OFFLINE == $iReturnCode )
         {
           $iOfflineCount++;
         }
         elsif ( VCSDef_MonitorExit::UNKNOWN == $iReturnCode )
         {
           $iUnknownCount++;
         }
  }


  if ( ( ( $iOnlineCount > 0 ) && ( $iOfflineCount > 0 ) ) || ( $iPartialOnlineCount > 0 ) )
  {
   @liCopyProcessPids = (@liCopyProcessPids, @liCopyPidFilePids);
   foreach my $iTmpPid (@liCopyProcessPids)
   {
     if (kill 0, $iTmpPid)
     {
       &logInfo ( 20077 , "<%s>: Sending SIGTERM to pid [%s]",
                           $sMyName , $iTmpPid);
       #-------------------------------
       # Send TERM in the first round.
       #-------------------------------
       kill 15, $iTmpPid;
     }
   }
   foreach my $iTmpPid (@liCopyProcessPids)
   {
     if (kill 0, $iTmpPid)
     {
       &logInfo ( 20078 ,"<%s>: Sending SIGKILL to pid [%s]",
                          $sMyName , $iTmpPid );
       #----------------------------------------------------------
       # Send KILL if the processes are still found to be running
       #----------------------------------------------------------
       kill 9, $iTmpPid;
     }
   }


  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCommand,
                                                 $sCommandArgs,
                                                 \$iExitCode,
                                                 $sUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )

  {
    #----------------------
    # Clean program success
    #----------------------
    &logTrace ( 20079, "<%s>: Clean command [%s] with args [%s] was successful",
                        $sMyName, $sCommand, $sCommandArgs);
  }
  else
  {
    if ( $bTimedOut )
    {
      #-------------------------
      # Clean program timed out.
      #-------------------------
      &logError ( 20080, "<%s>: Clean command [%s] with args [%s] timed out ",
                          $sMyName , $sCommand , $sCommandArgs);
    }
    else
    {
      #---------------------------------------------
      # Clean program did not execute successfully.
      #---------------------------------------------
      &logError ( 20081, "<%s>: Clean command [%s] with args [%s] failed to run",
                          $sMyName , $sCommand , $sCommandArgs);
    }

  }


   $iReturnCode = VCSDef_MonitorExit::OFFLINE;
  }

return $iReturnCode;
}

=pod

-------------------------------------------------------------------------
Function : int FFLMStatus ( IN String PidFile,
                            IN String Proc,
                            OUT RefInt Pid );

Description: Checks status of Process listed in PidFile, if running

Parameters :

$sPidFile       Absolute path to PID file of service
$sProc          Process string to search for in proc table
$riPid          Reference to existing PID

Returns : Integer value as defined below:
100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
-------------------------------------------------------------------------

=cut

sub FFLMStatus {
#----------------
# Get Arguments..
#----------------
my ( $sPidFile,
     $sProc,
     $riPid ) = @_;

#-----------------
# Announce Entry..
#-----------------
my $sMyName = ( caller 0 )[3];

&logTrace ( 20082, "Subroutine <%s> called with:
PidFile                [%s]
Proc                   [%s]
riPid                  [%s]",

$sMyName, $sPidFile, $sProc, $riPid );

#-----------------------------
# Set return code to success..
#-----------------------------
my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
$$riPid = 0; # Reset..

#-------------------------
# Define local variables..
#-------------------------
my ( $iPid, $rhlProc, @lProcessHash ) = ();

#----------------------------------
# Start by testing the attributes..
#----------------------------------
if ( $sProc !~ /\S/ ) {
  &logError ( 20083, "<%s> Called with empty [%s] process string",
  $sMyName, $sProc );
  $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
}
elsif ( ! -e $sPidFile ) {
  &logTrace ( 20084, "<%s> PID file [%s] does not exist",
  $sMyName, $sPidFile );
  $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
}
elsif ( ! -s $sPidFile ) {
  &logWarn ( 20085, "<%s> PID file [%s] is of zero bytes",
  $sMyName, $sPidFile );
  $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
}

#---------------------
# Return, if offline..
#---------------------
return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;


if ( open ( PIDFILE, $sPidFile ) ) {
  ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                       s/.*?(\d+)$/$1/  # Get PID..
                   } <PIDFILE>;

  $iPid = '' unless defined $iPid;      # Avoid undefs..

  close ( PIDFILE ) ||                  # Close file..
  &logWarn ( 20086,
  "<%s> Could not close PID file [%s] after read [%s]",
  $sMyName, $sPidFile, $! );
}
else {
  &logError ( 20087,
  "<%s> Failed to open PID file [%s] for read: [%s]",
  $sMyName, $sPidFile, $! );
  return VCSDef_MonitorExit::OFFLINE; # Failure..
}

#------------------------
# Check if PID is valid..
#------------------------
if ( $iPid =~ /\d+/ &&     # If the valid number..
     kill ( 0, $iPid ) ) { # is a PID that's alive..

  &logTrace ( 20088,
  "<%s> PID [%s] from PID file [%s] exists",
  $sMyName, $iPid, $sPidFile );

  #-----------------------
  # Search in proc table..
  #-----------------------
  if ( AgentBuilder::GetPidDetails ( $iPid,
                                     \@lProcessHash ) ) {
    #--------------
    # Got details..
    #--------------
    foreach $rhlProc ( @lProcessHash ) {
      &logTrace ( 20089,
      "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      #--------------
      # Now compare..
      #--------------
      if (  $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
        &logTrace ( 20090,
        "<%s> PID [%s]  with ARGS [%s] contains [%s]",
        $sMyName, $iPid, $$rhlProc { COMMAND }, $sProc );
        $$riPid = $iPid; # Store..
      }
      else {
        &logTrace ( 20091,
        "<%s> Ignoring PID [%s]  with ARGS [%s] with string [%s]",
        $sMyName, $iPid,  $$rhlProc { COMMAND }, $sProc );
        $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
      }
    }
  }
  else {
    &logTrace ( 20092,
    "<%s> Could not get details for PID [%s]",
    $sMyName, $iPid );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
}
else {
  &logTrace ( 20093, "<%s> PID [$iPid] does not exist",
  $sMyName, $sPidFile );
  $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
}

@lProcessHash = (); # Truncate..
return $iReturnCode;
}



=pod

-------------------------------------------------------------------------
Function : bool CreatePidFile ( IN String PidFile,
                                IN Integer Pid );


Description : Creates PID file with given PID

Parameters :

$sPidFile     File to create
$iPid         PID to store

Returns : boolean value in the set of {0,1} where
          0 false
          1 true
-------------------------------------------------------------------------

=cut

sub CreatePidFile {
#----------------
# Get Arguments..
#----------------
my ( $sPidFile,
     $iPid ) = @_;

#-----------------
# Announce Entry..
#-----------------
my $sMyName = ( caller 0 )[3];

&logTrace ( 20094, "Subroutine <%s> called with:
PidFile         [%s]
Pid             [%s]",

$sMyName, $sPidFile, $iPid );

#--------------------------------------
# Set Return Code Optimistically to 1..
#--------------------------------------
my $iReturnCode = 1;

#-----------------------
# Check against undefs..
#-----------------------
unless ( defined $iPid && $iPid =~ /^\d+$/o) {
  $iPid ||= 'undef'; # Prevent undefs.
  &logError ( 20095, "<%s> PID [%s] is not a valid positive number",
  $sMyName, $iPid );
  return 0; # Proceed no further..
}

#------------------
# Create PID file..
#------------------
&logTrace ( 20096,
"<%s> Creating PID file [%s] with PID [%s]",
$sMyName, $sPidFile, $iPid );

if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
               0600 ) ) {
  print PIDFILE $iPid, "\n"; # Update..

  unless ( close ( PIDFILE ) ) {
    &logWarn ( 20097,
    "<%s> Failed to close PidFile [%s] after update: [%s]",
    $sMyName, $sPidFile, $! );
    $iReturnCode = 0;
  }
}
else {
  &logWarn ( 20098,
  "<%s> Failed to create PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );
  $iReturnCode = 0;
}

return $iReturnCode;
}





=pod

-------------------------------------------------------------------------
Function : bool GetApplicationDetails ( IN RefArrayString MonitorProcessPatterns,
                                        OUT RefHashList AppDetails );

Description : Gives PID file and process string for given server type

Parameters  :

$arMonitorProcessPatterns  Store all process patterns to be monitored
$rhlAppDetails            Stores details of PID file and process string

Returns : boolean value in the set of {0,1} where
          0 false
          1 true
-------------------------------------------------------------------------

=cut


sub GetApplicationDetails 
{
#----------------
# Get Arguments..
#----------------
my ( $arMonitorProcessPatterns,
     $rhlAppDetails) = @_;

#-----------------
# Announce Entry..
#-----------------
my $sMyName = ( caller 0 )[3];

&logTrace( 20099, "Subroutine <%s> called with:
rhlAppDetails [%s]",
$sMyName, $rhlAppDetails );

foreach  my $key ( @$arMonitorProcessPatterns)
{
  &logTrace ( 20100, " <%s> MonitorProcessPatterns          [%s]", $sMyName, $key );
}


#-----------------------------
# Set return code to success..
#-----------------------------
my $iReturnCode = 1; # Set to true..

#-------------------
# Define variables..
#-------------------
my ( $sProc, $sPidFile, $sPIDDir, $sResourceType, $sResourceName, $iPat, $sKey ) = ();
my @lsSrchPatterns = @$arMonitorProcessPatterns;
#----------------------------------------------
# Derive our resource type and resource names..
#----------------------------------------------
$sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
$sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
$sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

#------------------------------
# Create PID folder if needed..
#------------------------------
if ( -d $sPIDDir ) {
  &logTrace ( 20101, "<%s> PID folder [%s] exists",
  $sMyName, $sPIDDir );
}
elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
  &logWarn ( 20102, "<%s> Failed to create PID folder [%s]",
  $sMyName, $sPIDDir );
  return 0; # Cannot proceed..
}

%$rhlAppDetails = (); # Truncate..

#-----------------------------------------------
# Set appropriate filter strings and PID files..
#-----------------------------------------------
foreach $sProc ( @lsSrchPatterns ) {
  $sKey = 'srchstr_' . ++ $iPat;
  $sPidFile = "$sPIDDir/pidfile.$sKey";
  @{ $$rhlAppDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
}

return $iReturnCode;
}

=pod

-------------------------------------------------------------------------
Function : int Monitor ( IN String  MonitorProgram
                         IN Int     MonitorProgramFrequency
                         IN String  User
                         IN String  EnvFile
                         IN Int     RelevantTimeout);


Parameters :

$sMonitorProgram             External Monitor program for the AgentBuilder application
$iMonitorProgramFrequency    Cycle count to run monitor program
$sUser                       AgentBuilder application user
$sEnvFile                    Environment File
$iRelevantTimeout            Timeout for Entry point

Returns : Integer value as defined below:

100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
-------------------------------------------------------------------------

=cut

sub Monitor {
#----------------
# Get Arguments..
#----------------
my (  $sMonitorProgram,
      $iMonitorProgramFrequency,
      $sUser,
      $sEnvFile,
      $iRelevantTimeout  ) = @_;

#-----------------
# Announce Entry..
#-----------------
my $sMyName = ( caller 0 )[3];

&logTrace ( 20103, "Subroutine <%s> called with:
  MonitorProgram                [%s]
  MonitorProgramFrequency       [%s]
  User                          [%s]
  EnvFile                       [%s]
  RelevantTimeout               [%s]",
  $sMyName,
  $sMonitorProgram,
  $iMonitorProgramFrequency,
  $sUser,
  $sEnvFile,
  $iRelevantTimeout);

#-----------------------------
# Set return code to success..
#-----------------------------
my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..


#----------------------------------------------
# In-depth monitoring, if correct time to run..
#----------------------------------------------

  #------------------------------------------
  # Determine the expiration time of the EP.
  #------------------------------------------
  my ($sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  my $iTillWhen = undef;


  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iMonitorProgramFrequency ) )
  {
    $iTillWhen = $^T + AgentBuilder::EPOP;
    &logTrace ( 20104," <%s> Proceeding with monitor program", $sMyName );
    if ( $sMonitorProgram )
    {
       $iRelevantTimeout = GetEPTimeout($sEntryPointName, $sResourceType);
       if ( $iRelevantTimeout <= 0 )
       {
         &logError( 20105, "<%s>: failed to determine entry point timeout,defaulting to 60", $sMyName);
         $iRelevantTimeout = 60;
       }
       $iTillWhen = ( $^T + $iRelevantTimeout ) - AgentBuilder::EPOP;
    }

    #---------------------------------------
    # MonitorProgram defined..
    # Separate the executable and the args..
    #---------------------------------------
    my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram         # Separate
                                  =~ /(\S+)\s*(.*)/;         # Cmd and args..

    $sExtCmdArgs = '' unless defined $sExtCmdArgs;           # Prevent undefs..

    if ( $sExtCmd ) {                                        # Defined..
       if ( VCSDef_MonitorExit::ONLINE ==                    # and found online..
                                           ( $iReturnCode =  # by the status..
                                                             AgentBuilder::ExternalMonitor ( $sExtCmd,
                                                                                             $sExtCmdArgs,
                                                                                             $iTillWhen,
                                                                                             $sUser,
                                                                                             $sEnvFile ) ) ) {
               #--------------------------------------------
               # LOG TRACE MESSAGE THAT INSTANCE IS ONLINE..
               #--------------------------------------------
               &logTrace ( 20106, "<%s>: External Monitor Program -> Instance is online ",$sMyName);
       }
       else {
               #------------------------------------------------
               # LOG ERROR MESSAGE THAT INSTANCE IS NOT ONLINE..
               #------------------------------------------------
               &logTrace ( 20107, "<%s>: External Monitor Program -> Instance is not online ",$sMyName);
       }
    }
  }

  return $iReturnCode;
}




=pod
      
-------------------------------------------------------------------------
Function : bool GetPidDetails ( IN Integer Pid, 
                                OUT RefList ProcList );

Description : Gets long process details for a known Pid if running

Parameters :

$iPid         PID whose details are requested
$rlProcList   Ref to List to store details

Returns : boolean value in the set of {0,1} where
          0 false
          1 true
-------------------------------------------------------------------------

=cut

sub GetPidDetails {
#----------------
# Get Arguments..
#----------------
my ( $iPid,
     $rlProcList ) = @_;

#-----------------
# Announce Entry..
#-----------------
my $sMyName = ( caller 0 )[3];

&logTrace ( 20108, "Subroutine <%s> called with:
Pid             [%s]
rlProcList      [%s]",

$sMyName, $iPid, $rlProcList );

#--------------------------------------
# Set Return Code Optimistically to 1..
#--------------------------------------
my $iReturnCode = 1;

#---------------
# Validate PID..
#---------------
unless ( defined $iPid    &&   # Catch undefs..
         $iPid =~ /^\d+$/ &&   # And invalid PIDs..
         kill ( 0, $iPid ) ) { # And defunct ones..
&logError ( 20109,
                  "<%s> PID [%s] is invalid", $sMyName, $iPid );
return 0;
}

#-------------------------
# Define local variables..
#-------------------------
my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
     %hsProc ) = ();

$sSwitch = $^O;    # Get OS..
@$rlProcList = (); # Truncate..

#-------------
# Store LANG..
#-------------
$sLang = $ENV { LANG };
$ENV { LANG } = 'C';

#---------------------------
# Switch to the correct OS..
#---------------------------
goto $sSwitch; {
  solaris:

    $sPsCmd = "/usr/ucb/ps";
    $sPsCmdArgs = "axwwl $iPid";
    $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  last;

  hpux:

    #-----------------------
    # Force XPG4 behaviour..
    #-----------------------
    $sUnix95 = $ENV { UNIX95 };
    $ENV { UNIX95 } = 'XPG4';

    $sPsCmd = "/usr/bin/ps";
    $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
    $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;

    aix:

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

    last;


    linux:

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

    last;
  }

  &logTrace ( 20110, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20111,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20112,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}





=pod
 -------------------------------------------------------------------------
 Function : int GetEPTimeout ( IN String EntryPointName
                               IN String ResourceType  );

 Description : Returns the Entry Point Timeout for the EP which we are in

 Parameters :

 $sEntryPointName           The EP which we are currently running
 $sResourceType             VCS resource type

 Returns  : Integer value as defined below:

 -------------------------------------------------------------------------
=cut


sub GetEPTimeout
{
  my ( $sEntryPointName, 
       $sResourceType) = @_;

  #------------------
  # Announce Entry..
  #------------------

  my $sMyName = ( caller 0 )[3];
  &logTrace( 20113, "Subroutine <%s> called with:
  EntryPointName [%s]
  ResourceType   [%s]",
  $sMyName , $sEntryPointName , $sResourceType );

  my $sObjectName          = "hatype";
  my %hTypeValues;

  my $iEngineTimeout;
  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20149,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
    return ( $iEngineTimeout ); 
  }

  if ( ! VCSagentFW::GetVCSObjectHash($sObjectName,$sResourceType,\%hTypeValues) )
  {
    &logError(20114,"<%s>:GetVCSObjectHash call failed.",$sMyName);
    return 0;
  }

  #-------------------------------
  # Make "monitor" to "Monitor"
  # offline to Offline and so on
  #-------------------------------

  my $sTempEntry =  $sEntryPointName ;
  $sTempEntry    =~ s/(\w)(.*)$/\U$1\L$2/;

  &logTrace(20115,"<%s>: Returning [%s] value [%s]", $sMyName, $sTempEntry."Timeout",$hTypeValues{$sTempEntry."Timeout"});
  return ( $hTypeValues{$sTempEntry."Timeout"});
}

=pod

 -------------------------------------------------------------------------
 Function : int SocketTest( IN String  User,
                            IN String  EnvFile,
                            IN String  Host,
                            IN String  Port);

 Description : Does in-depth monitoring, if enabled.

 Parameters :
 $sUser                       AgentBuilder application user
 $sEnvFile                    Environment File 
 $sHost                       Hostname or IP address
 $sPort                       Port number for socket test
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SocketTest {
  #----------------
  # Get Arguments..
  #----------------
  my (  $sUser,
        $sEnvFile,
        $sHost,
        $sPort ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20116, "Subroutine <%s> called with:
  Host                 [%s]
  Port                 [%s]",

  $sMyName, $sHost, $sPort );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;


  if (  Comms::CheckPort($sHost,$sPort) )
  {
    &logTrace(20117, "<%s>: Server responding at [%s:%s] ", $sMyName,$sHost,$sPort );
  $iReturnCode = VCSDef_MonitorExit::ONLINE;
  }
  else
  {
    &logTrace(20118, "<%s>: Server not responding at [%s:%s] ",
                    $sMyName,$sHost,$sPort );
  $iReturnCode = VCSDef_MonitorExit::OFFLINE;
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen,
                                  IN String  User,
                                  IN String  EnvFile);

 Description : Handles user defined external monitor program

 Parameters :

 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 $sUser                       AgentBuilder application user
 $sEnvFile                    Environment File 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my (  $sCmd,
        $sCmdArgs,
        $iTillWhen,
        $sUser,
        $sEnvFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20119, "Subroutine <%s> called with:
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]
  User                     [%s]
  EnvFile                  [%s]",
  $sMyName, $sCmd, $sCmdArgs, $iTillWhen, $sUser, $sEnvFile );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeOut, @lsCmdOut ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20120,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20121,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeOut = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20122,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeOut ) );
   &logError ( 20123,
    "<%s> time=[%s] iBuf [%s] external monitor [%s].
    Short by [%s] sec", $sMyName, time, $iBuf, $sCmd, abs ( $iTimeOut ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20124,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeOut );

  #------------------------------------------
  # EXECUTE THE COMMAND USING F/W FUNCTIONS..
  #------------------------------------------
  my $bSourceEnv = 0;
  my @lsCommandOut = ();
  my $bTimedOut;
  my $sOutFileName = undef;
  my $sFromDir = undef;

  if ( Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                 $bSourceEnv,
                                                 $sCmd,
                                                 $sCmdArgs,
                                                 \$iExitCode,
                                                 $sUser,
                                                 \@lsCommandOut,
                                                 \$bTimedOut,
                                                 $iTimeOut,
                                                 $sOutFileName,
                                                 $sFromDir      ) )
   {
   
  #------------------------
  # Report on exit status..
  #------------------------
    if ( 0 == $iExitCode || VCSDef_MonitorExit::ONLINE == $iExitCode )
    {
      #------------------------------------------------------------
      # External MonitorProgram reported  resource to be ONLINE
      #------------------------------------------------------------
      &logTrace ( 20125, "<%s>: MonitorProgram [%s] with arguments [%s] reporting resource ONLINE, exit code [%s]",
                  $sMyName , $sCmd , $sCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::ONLINE;
    }
    elsif ( 1 == $iExitCode || VCSDef_MonitorExit::OFFLINE == $iExitCode )
    {
      #------------------------------------------------------------
      # External MonitorProgram reported resource to be OFFLINE
      #------------------------------------------------------------
      &logTrace ( 20126, "<%s>: MonitorProgram [%s] with arguments [%s] reporting resource OFFLINE, exit code [%s]",
                  $sMyName , $sCmd , $sCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    else
    {
      #------------------------------------------------------------
      # External MonitorProgram reported resource to be UNKNOWN
      #------------------------------------------------------------
      &logError ( 20127, "<%s>: MonitorProgram [%s] with arguments [%s] reporting resource UNKNOWN, exit code [%s]",
                         $sMyName , $sCmd , $sCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
  }
  else
  {
    if ( $bTimedOut )
    {
      #----------------------------------------------------------------------
      # Command timed out!  Return UNKNOWN as the first (and possibly second)
      # level monitoring did work -- it is considered safer to return UNKNOWN
      # as the service may be generally OK, and the MonitorProgram didn't
      # determine for sure that the service was not up, as the defined
      # methodology to do so (as implemented in the MonitorProgram) could not
      # be run successfully (it timed out).
      #----------------------------------------------------------------------
      &logError ( 20128, "<%s>: MonitorProgram [%s] with arguments [%s] timed out",
                  $sMyName , $sCmd , $sCmdArgs );
      &logTrace ( 20129, "<%s>: MonitorProgram is not responding, sleeping till _had kills the entry point.",
                  $sMyName );
      while (1) {sleep 1000};

    }
    else
    {
      &logError ( 20129, "<%s>: MonitorProgram [%s] with arguments [%s] did not execute successfully, exit code[%s]",
                  $sMyName , $sCmd , $sCmdArgs , $iExitCode);
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
  }



  if ( ! defined $iExitCode ) {
    &logError ( 20130,
    "<%s> Command [%s] with args [%s] timed out",
    $sMyName, $sCmd, $sCmdArgs );
    $iExitCode = 'undef';
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20131,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }
  else {
    &logTrace ( 20132,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }


  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;
  return $iReturnCode;
}

1; # So PERL knows its included..
